# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from collections import namedtuple
from datetime import timezone, timedelta
from functools import wraps
from datetime import datetime
import os
import re

from msprobe.core.common.const import MonitorConst
from msprobe.core.common.log import logger
from msprobe.core.common.utils import is_int
from msprobe.core.common.file_utils import check_file_or_directory_path, recursive_chmod

beijing_tz = timezone(timedelta(hours=8))
MVResult = namedtuple('MVResult', ("exp_avg", "exp_avg_sq", "update", "ratio"))


class MsgConst:
    """
    Class for log messages const
    """
    SPECIAL_CHAR = ["\n", "\r", "\u007F", "\b", "\f", "\t", "\u000B", "%08", "%0a", "%0b", "%0c", "%0d", "%7f"]


def get_output_base_dir():
    return os.getenv(MonitorConst.MONITOR_OUTPUT_DIR, MonitorConst.DEFAULT_MONITOR_OUTPUT_DIR)


def filter_special_chars(func):
    @wraps(func)
    def func_level(msg):
        for char in MsgConst.SPECIAL_CHAR:
            msg = msg.replace(char, '_')
        return func(msg)

    return func_level


def validate_ops(ops):
    if not isinstance(ops, list):
        raise TypeError("ops should be a list")
    valid_ops = []
    for op in ops:
        if op not in MonitorConst.OP_LIST:
            logger.warning(f"op {op} is not supported. Optional ops: {MonitorConst.OP_LIST}")
            continue
        valid_ops.append(op)
    if not valid_ops:
        default_op = MonitorConst.OP_LIST[0]
        valid_ops.append(default_op)
        logger.info_on_rank_0(f"There is no valid ops, default op {default_op} is used")
    # 增加默认shape和dtype参数
    if "shape" not in valid_ops:
        valid_ops.append("shape")
    if "dtype" not in valid_ops:
        valid_ops.append("dtype")
    return valid_ops


def validate_ndigits(ndigits):
    if not ndigits:
        return
    if not is_int(ndigits) or ndigits <= 0:
        raise ValueError(f"ndigits({ndigits}) is not a positive integer, current is: {ndigits}.")
    if ndigits > MonitorConst.MAX_NDIGITS:
        raise ValueError(f"The maximum supported ndigits is {MonitorConst.MAX_NDIGITS}, current value: {ndigits}.")


def validate_ranks(ranks):
    if not isinstance(ranks, list):
        raise TypeError("module_ranks should be a list")
    for rank in ranks:
        if not isinstance(rank, int) or isinstance(rank, bool):
            raise TypeError(f"element in module_ranks should be a int, get {type(rank)}")


def validate_targets(targets):
    if not isinstance(targets, dict):
        raise TypeError('targets in config.json should be a dict')
    for module_name, field in targets.items():
        if not isinstance(module_name, str):
            raise TypeError('key of targets should be module_name[str] in config.json')
        if not isinstance(field, dict):
            raise TypeError('values of targets should be cared filed e.g. {"input": "tensor"} in config.json')


def validate_l2_targets(targets):
    if not isinstance(targets, dict):
        raise TypeError('l2_targets in config.json should be a dict')
    for hook_name, target_list in targets.items():
        if hook_name not in MonitorConst.L2_HOOKS:
            raise TypeError(f'key of l2_targtes must be in {MonitorConst.L2_HOOKS}, got {hook_name}')
        if not isinstance(target_list, list):
            raise TypeError('values of l2_targets should be a list in config.json')
        for item in target_list:
            if not isinstance(item, str):
                raise TypeError(f'item of "{hook_name}" in l2_targets should be module_name[str] in config.json')


def validate_recording_l2_features(recording_l2_features):
    if not isinstance(recording_l2_features, bool):
        raise TypeError("recording_l2_features should be a bool")


def validate_sa_order(sa_order):
    if isinstance(sa_order, str):
        sa_order = sa_order.replace(' ', '')
    if sa_order not in MonitorConst.SA_ORDERS:
        raise TypeError(f'sa_order must be in {MonitorConst.SA_ORDERS}, got {sa_order}')


def validate_print_struct(print_struct):
    if not isinstance(print_struct, bool):
        raise TypeError("print_struct should be a bool")


def validate_ur_distribution(ur_distribution):
    if not isinstance(ur_distribution, bool):
        raise TypeError('ur_distribution should be a bool')


def validate_xy_distribution(xy_distribution):
    if not isinstance(xy_distribution, bool):
        raise TypeError('xy_distribution should be a bool')


def validate_wg_distribution(wg_distribution):
    if not isinstance(wg_distribution, bool):
        raise TypeError('wg_distribution should be a bool')


def validate_mg_distribution(mg_distribution):
    if not isinstance(mg_distribution, bool):
        raise TypeError('mg_distribution should be a bool')


def validate_param_distribution(param_distribution):
    if not isinstance(param_distribution, bool):
        raise TypeError('param_distribution should be a bool')


def validate_cc_distribution(cc_distribution):
    if not isinstance(cc_distribution, dict):
        raise TypeError('cc_distribution should be a dictionary')
    for key, value in cc_distribution.items():
        if key == 'enable':
            if not isinstance(value, bool):
                raise TypeError('cc_distribution enable should be a bool')
        elif key == 'cc_codeline':
            if not isinstance(value, list):
                raise TypeError('cc_distribution cc_codeline should be a list')
        elif key == 'cc_pre_hook':
            if not isinstance(value, bool):
                raise TypeError('cc_distribution cc_pre_hook should be a bool')
        elif key == 'cc_log_only':
            if not isinstance(value, bool):
                raise TypeError('cc_distribution cc_log_only should be a bool')
        else:
            raise TypeError(f'{key} of cc_distribution is not supported.')


def validate_squash_name(squash_name):
    if not isinstance(squash_name, bool):
        raise TypeError('squash_name should be a bool')


def validate_alert(alert):
    if not isinstance(alert, dict):
        raise TypeError('alert should be a dictionary')
    rules = alert.get('rules')
    if rules and isinstance(rules, list):
        for rule in rules:
            rule_name = rule.get("rule_name")
            if rule_name and rule_name not in MonitorConst.RULE_NAME:
                raise TypeError(f"{rule_name} is not supported")
            args = rule.get("args")
            if args and isinstance(args, dict):
                threshold = args.get("threshold")
                if not isinstance(threshold, (float, int)) or threshold < 0:
                    raise TypeError('threshold must be float and not less than 0')
    dump = alert.get('dump')
    if dump and not isinstance(dump, bool):
        raise TypeError('dump must be bool.')


def validate_step_count_per_record(step_count_per_record):
    if not is_int(step_count_per_record):
        raise TypeError('step_count_per_record must be int.')
    if step_count_per_record < 1:
        raise ValueError("step_count_per_record must greater than 0")
    if step_count_per_record > 1e6:
        raise ValueError("step_count_per_record must smaller than 1e6")


def validate_dynamic_on(dynamic_on):
    if not isinstance(dynamic_on, bool):
        raise TypeError('dynamic_on should be a bool')


def validate_monitor_mbs_grad(monitor_mbs_grad):
    if not isinstance(monitor_mbs_grad, bool):
        logger.warning(f'monitor_mbs_grad should be a bool, actual value is {monitor_mbs_grad}.')
        return False
    return monitor_mbs_grad


def validate_append_output(append_output):
    if not isinstance(append_output, list):
        raise TypeError('append_output should be a list')
    if len(append_output) > 0 and len(append_output) != 2:
        raise ValueError('append_output should be empty or contain exactly 2 elements')


def validate_config(config):
    config['ops'] = validate_ops(config.get('ops', []))

    ndigits = config.get('ndigits')
    validate_ndigits(ndigits)

    eps = config.get('eps', 1e-8)
    if not isinstance(eps, float):
        raise TypeError("eps should be a float")

    ranks = config.get("module_ranks", [])
    validate_ranks(ranks)

    targets = config.get("targets", {})
    validate_targets(targets)

    l2_targets = config.get("l2_targets", {})
    validate_l2_targets(l2_targets)

    recording_l2_features = config.get("recording_l2_features", False)
    validate_recording_l2_features(recording_l2_features)

    sa_order = config.get("sa_order", "s,b,h,d")
    validate_sa_order(sa_order)

    print_struct = config.get('print_struct', False)
    validate_print_struct(print_struct)

    ur_distribution = config.get('ur_distribution', False)
    validate_ur_distribution(ur_distribution)

    xy_distribution = config.get('xy_distribution', False)
    validate_xy_distribution(xy_distribution)

    wg_distribution = config.get('wg_distribution', False)
    validate_wg_distribution(wg_distribution)

    mg_distribution = config.get('mg_distribution', False)
    validate_mg_distribution(mg_distribution)

    param_distribution = config.get('param_distribution', False)
    validate_param_distribution(param_distribution)

    cc_distribution = config.get('cc_distribution', {})
    validate_cc_distribution(cc_distribution)

    alert = config.get('alert', {})
    validate_alert(alert)

    step_count_per_record = config.get('step_count_per_record', 1)
    validate_step_count_per_record(step_count_per_record)

    config["start_step"] = validate_int_arg(config.get("start_step"), "start_step",
                                            MonitorConst.DEFAULT_START_STEP, MonitorConst.DEFAULT_START_STEP)
    config["collect_times"] = validate_int_arg(config.get("collect_times"), "collect_times",
                                               MonitorConst.DEFAULT_MIN_COLLECT_TIMES,
                                               MonitorConst.DEFAULT_MAX_COLLECT_TIMES)
    config["step_interval"] = validate_int_arg(config.get("step_interval"), "step_interval",
                                               MonitorConst.DEFAULT_STEP_INTERVAL, MonitorConst.DEFAULT_STEP_INTERVAL)

    squash_name = config.get('squash_name', True)
    validate_squash_name(squash_name)

    time_tags = config.get("append_output", [])
    validate_append_output(time_tags)

    config["monitor_mbs_grad"] = validate_monitor_mbs_grad(config.get('monitor_mbs_grad', False))

    dynamic_on = config.get('dynamic_on', False)
    validate_dynamic_on(dynamic_on)

    if not targets:
        if xy_distribution:
            config["all_xy"] = True
        config["targets"] = {"": {}}


def time_str2time_digit(time_str):
    time_format = '%b%d_%H-%M-%S'
    if not isinstance(time_str, str):
        raise TypeError(f"time_str:{time_str} should be a str")
    try:
        time_digit = datetime.strptime(time_str, time_format)
    except Exception as e:
        raise RuntimeError(f"illegal timestamp: {time_str}, timestamp should be prefix \
                           of existing output dirpath, like 'Dec03_21-34-40'.") from e
    return time_digit


def get_target_output_dir(monitor_path, time_start, time_end):
    check_file_or_directory_path(monitor_path, isdir=True)
    time_start = time_str2time_digit(time_start) if time_start is not None else time_start
    time_end = time_str2time_digit(time_end) if time_end is not None else time_end
    if time_start and time_end and time_start > time_end:
        raise ValueError(f"time_start({time_start}) greater than time_end({time_end})")
    result = {}
    for dirname in os.listdir(monitor_path):
        match = re.match(MonitorConst.OUTPUT_DIR_PATTERN, dirname)
        if not match:
            continue
        time_tag = match.group(1)
        rank = match.group(2)
        target_time = time_str2time_digit(time_tag)
        start_ok = time_start is None or target_time >= time_start
        end_ok = time_end is None or target_time <= time_end
        if start_ok and end_ok:
            result[rank] = os.path.join(monitor_path, dirname)
    return result


def chmod_tensorboard_dir(path):
    """
        format配置为tensorboard时，需要补充文件权限设置
    """
    try:
        recursive_chmod(path)
    except Exception as e:
        logger.warning(f"chmod tensorboard dir wrong because {e}, not updated, please check!!!")


def validate_set_monitor(grad_acc_steps, start_iteration):
    """
    validate parameters of set_monitor.
    """
    grad_acc_steps = validate_int_arg(grad_acc_steps, "grad_acc_steps",
                                      MonitorConst.DEFAULT_GRAD_ACC_STEPS, MonitorConst.DEFAULT_GRAD_ACC_STEPS)

    start_iteration = validate_int_arg(start_iteration, "start_iteration",
                                       MonitorConst.DEFAULT_START_ITERATION, MonitorConst.DEFAULT_START_ITERATION)
    return grad_acc_steps, start_iteration


def validate_int_arg(value, name, minimum, default_value):
    """Validate int args, if any exception occurs, use the default value."""
    if value is None:
        return default_value
    try:
        if not is_int(value):
            raise TypeError(f"{name} must be int")
        if value < minimum:
            raise ValueError(f"{name} must greater than {minimum}")
    except Exception as e:
        value = default_value
        logger.warning(f"Validate {name} failed, {e}, replaced with default value {value}.")
    return value



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import gc
import sys
from pathlib import Path
import mindspore
from msprobe.mindspore.common.log import logger
from msprobe.core.common.const import Const, CompareConst
from msprobe.mindspore.common.const import MsCompareConst
import torch as mindtorch
from torch import Tensor as mindtorch_tensor
import torch.nn.functional as mindtorch_func
import torch.distributed as mindtorch_dist


is_valid_pt_mt_env = True


def is_mindtorch():
    mindtorch_check_result = False
    try:
        import torch as test_torch
        from mindspore import Tensor as MindsporeTensor
    except ImportError:
        return mindtorch_check_result
    tensor = test_torch.tensor(0.0)
    if isinstance(tensor, MindsporeTensor):
        mindtorch_check_result = True

    return mindtorch_check_result


def remove_torch_related_paths():
    removed_paths = []
    if not is_mindtorch():
        return
    try:
        import torch as remove_torch
        torch_file = remove_torch.__file__
    except ImportError:
        return

    torch_dir = os.path.dirname(torch_file)

    torch_dir_path = Path(torch_dir).resolve()
    parent_dir = torch_dir_path.parent

    paths_to_remove = [str(parent_dir)]

    for path in paths_to_remove:
        try:
            path_resolved = str(Path(path).resolve())
        except Exception as error:
            logger.debug(f"Failed to resolve path {path}: {error}")
            continue

        if path_resolved in sys.path:
            index = sys.path.index(path_resolved)
            removed_paths.append((path_resolved, index))
            sys.path.pop(index)

    return


def clear_torch_from_sys_modules():
    modules_to_remove = []
    for module in sys.modules:
        if module == "torch" or module.startswith("torch."):
            modules_to_remove.append(module)

    for module in modules_to_remove:
        del sys.modules[module]


def set_pt_mt_env_invalid():
    global is_valid_pt_mt_env
    is_valid_pt_mt_env = False


def delete_torch_paths():

    if not is_mindtorch():
        set_pt_mt_env_invalid()

    clear_torch_from_sys_modules()

    for count_delete_env_path in range(MsCompareConst.MAX_RECURSION_DEPTH):
        if not is_mindtorch():
            break

        remove_torch_related_paths()

        clear_torch_from_sys_modules()

        if count_delete_env_path >= MsCompareConst.MAX_RECURSION_DEPTH - 1:
            raise Exception(f"Please check if you have a valid PyTorch and MindTorch environment, and ensure "
                            f"the PYTHONPATH environment variable depth does not "
                            f"exceed {MsCompareConst.MAX_RECURSION_DEPTH}.")


if not is_mindtorch():
    set_pt_mt_env_invalid()

else:
    initial_sys_path = sys.path.copy()
    delete_torch_paths()

    gc.collect()

    import torch

    if is_mindtorch():
        set_pt_mt_env_invalid()

    sys.path = initial_sys_path





# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the overflow advisor function.
"""

from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.advisor.advisor_const import AdvisorConst
from msprobe.msaccucmp.advisor.advisor_result import AdvisorResult


class OverflowAdvisor:
    """
    Class for generate overflow advisor
    """

    def __init__(self, input_file, result):
        self.analyze_data = input_file
        self.result = result

    def start_analyze(self):
        """
        Analyze result by overflow detection
        """
        log.print_info_log('Start FP16 Overflow detection.')
        data_columns = self.analyze_data.columns.values
        if AdvisorConst.OVERFLOW not in data_columns:
            log.print_warn_log('Input csv file does not contain %s columns, Skip FP16 Overflow detection.'
                               % AdvisorConst.OVERFLOW)
        else:
            overflow_df = self.analyze_data[self.analyze_data[AdvisorConst.OVERFLOW] == "YES"]
            # check overflow dataframe lines
            if overflow_df.shape[0] == 0:
                log.print_info_log('After analysis, input csv file does not have FP16 Overflow problem.')
                return self.result
            overflow_df.reset_index(drop=True, inplace=True)
            index = overflow_df.at[0, AdvisorConst.INDEX]
            self.result = AdvisorResult(True, AdvisorConst.OVERFLOW_DETECTION, str(index),
                                        AdvisorConst.OVERFLOW_SUGGEST)
        return self.result





# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This class mainly involves the main function.
"""

import os
import re
import sys
import argparse
import time

from msprobe.msaccucmp.cmp_utils import log, path_check, file_utils
from msprobe.msaccucmp.cmp_utils.utils import safe_path_string, check_file_size
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.reg_manager import RegManager
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils.path_check import check_others_permission
from msprobe.msaccucmp.algorithm_manager.algorithm_manager import AlgorithmManagerMain
from msprobe.msaccucmp.compare_vector import VectorComparison
from msprobe.msaccucmp.conversion.shape_format_conversion import FormatConversionMain
from msprobe.msaccucmp.dump_parse.dump_data_parser import DumpDataParser
from msprobe.msaccucmp.overflow.overflow_analyse import OverflowAnalyse
from msprobe.msaccucmp.pytorch_cmp.compare_pytorch import PytorchComparison
from msprobe.msaccucmp.vector_cmp.batch_compare import BatchCompare


MIND_STUDIO_LOGO = "[Powered by MindStudio]"


def _get_algorithm_help_info() -> str:
    """
    get algorithm help info
    :return: help info
    """
    algorithm_help_info = ['<Optional> the algorithm selection, built-in algorithm where ']
    for index, item in enumerate(ConstManager.BUILT_IN_ALGORITHM):
        algorithm_help_info.append("".join([str(index), "=", item, ", "]))
    algorithm_help_info.append('The custom algorithm uses the algorithm name. '
                               'The selection splits by ",", such as "0,MaxAbsoluteError,4,CustomAlg" or "all". '
                               'The default value is "all"')
    return "".join(algorithm_help_info)


def _get_algorithm_help_info_for_int(default_value_list: list) -> str:
    """
    get algorithm help info for int
    :param default_value_list:
    :return:
    """
    algorithm_help_info = ['<Optional> the algorithm selection, ']
    for index, item in enumerate(ConstManager.BUILT_IN_ALGORITHM):
        algorithm_help_info.append("".join([str(index), ":", item, ", "]))
    default_info = "The default value is %s." % default_value_list
    algorithm_help_info.append(default_info)
    return "".join(algorithm_help_info)


def _match_built_in_arg_value(alg_arg: str) -> bool:
    # the alg value like -alg 1 2 3 --algorithm 1 4
    if alg_arg in sys.argv:
        alg_index = sys.argv.index(alg_arg)
        re_pattern = re.compile(RegManager.BUILTIN_ALGORITHM_INDEX_PATTERN)
        # check the second value after -alg match BUILTIN_ALGORITHM_INDEX_PATTERN
        if alg_index + 2 < len(sys.argv):
            if re_pattern.match(sys.argv[alg_index + 2]):
                return True
    return False


def _add_alg_argument(compare_parser: argparse.ArgumentParser) -> None:
    if _match_built_in_arg_value('-alg') or _match_built_in_arg_value('--algorithm'):
        default_value_list = []
        for index, _ in enumerate(ConstManager.BUILT_IN_ALGORITHM):
            default_value_list.append(index)
        compare_parser.add_argument(
            '-alg', '--algorithm', dest='algorithm', type=int, nargs="+", choices=default_value_list,
            default=default_value_list, help=_get_algorithm_help_info_for_int(default_value_list))
    else:
        compare_parser.add_argument(
            '-alg', '--algorithm', dest='algorithm', default="all", type=safe_path_string,
            help=_get_algorithm_help_info())


def _add_fusion_rule_argument(compare_parser: argparse.ArgumentParser) -> None:
    compare_parser.add_argument(
        '-f', '--fusion_rule_file', dest='fusion_rule_file', default='', type=safe_path_string,
        help='<Optional> the fusion rule file path')
    compare_parser.add_argument(
        '-q', '--quant_fusion_rule_file', dest='quant_fusion_rule_file', type=safe_path_string,
        default='', help='<Optional> the quant fusion rule file path')
    compare_parser.add_argument(
        '-cf', '--close_fusion_rule_file', dest='close_fusion_rule_file', type=safe_path_string,
        default='', help='<Optional> the rule file path without fusion')


def _compare_parser(compare_parser: argparse.ArgumentParser) -> None:
    compare_parser.add_argument(
        '-m', '--my_dump_path', dest='my_dump_path', default='', type=safe_path_string, required=True,
        help='<Required> my dump path, the data compared with golden data')
    compare_parser.add_argument(
        '-g', '--golden_dump_path', dest='golden_dump_path', default='', type=safe_path_string,
        help='<Required> the golden dump path', required=True)
    _add_fusion_rule_argument(compare_parser)
    compare_parser.add_argument(
        '-out', '--output', dest='output_path', type=safe_path_string, default='', help='<Optional> the output path')
    compare_parser.add_argument(
        '-c', '--custom_script_path', dest='custom_script_path', default='', type=safe_path_string,
        help='<Optional> the user-defined script path, including format conversion and algorithm')
    compare_parser.add_argument(
        '-a', '--algorithm_options', dest='algorithm_options', default='',
        help='<Optional> the arguments for each algorithm. The format is "algorithm_name:param_name='
             'param_value". The parameter splits by ",". The algorithm splits by ";". '
             'Such as "CosineSimilarity:max=1,min=0;aa:max=1,min=0"')
    _add_alg_argument(compare_parser)
    compare_parser.add_argument(
        '-map', '--mapping', dest="mapping", action="store_true", required=False,
        help="<Optional> create mappings between my output operators and ground truth one")
    compare_parser.add_argument(
        "-overflow_detection", dest="overflow_detection", action="store_true", required=False,
        help="<Optional> Operator overflow detection, only operators of the fp16 type are supported")
    compare_parser.add_argument(
        '-r', '--range', dest="range", default=None, required=False,
        help='<Optional> compare network with the range. The format is "start,end,step". '
             '`start` means the count starts position, limited to [1, op_count], default 1.'
             '`end` means the count ends position, limited to [>=start, op_count] or -1, default -1'
             '`step` limited to [1, op_count], default 1. -r command and -s command can not be used at the same time')
    compare_parser.add_argument(
        '-s', '--select', dest="select", default=None, required=False,
        help='<Optional> compare network with the range. The format is "index_1, index_2,..." Every index should be'
             'a number in the fusion operator list -r command and -s command can not be used at the same time')
    compare_parser.add_argument(
        '-p', '--post_process', dest='post_process', choices=[0, 1], type=int, default=None,
        help='<Optional> whether to extract the compare result, only pytorch is supported.'
             '0 indicates the comparison result is not extracted, 1 indicates the comparison result is extracted')

    compare_parser.add_argument(
        '-max', '--max_cmp_size', dest='max_cmp_size', type=int, default=0,
        help='<Optional> max size of tensor array to compare')

    _add_advisor_argument(compare_parser)
    _add_version_argument(compare_parser)
    _add_argument_for_single_op(compare_parser)
    _add_ffts_argument(compare_parser)


def _add_advisor_argument(compare_parser: argparse.ArgumentParser) -> None:
    compare_parser.add_argument(
        '-advisor', dest="advisor", action="store_true", required=False, help="<optional> Enable advisor after compare"
    )


def _add_ffts_argument(compare_parser: argparse.ArgumentParser) -> None:
    compare_parser.add_argument(
        '-ffts', dest="ffts", action="store_true",
        help="<optional> Enable the comparison between ffts+ and ffts+. "
             "Direct comparison is performed without data combination")


def _add_argument_for_single_op(compare_parser: argparse.ArgumentParser) -> None:
    compare_parser.add_argument('-op', '--op_name', dest='op_name', default=None, help='<Optional> operator name')
    group = compare_parser.add_mutually_exclusive_group()
    group.add_argument(
        '-o', '--output_tensor', dest='output', default=None,
        help='<Optional> the index of output, takes effect only when the "-op" exists')
    group.add_argument(
        '-i', '--input_tensor', dest='input', default=None,
        help='<Optional> the index for input, takes effect only when the "-op" exists')
    compare_parser.add_argument(
        '--ignore_single_op_result', dest="ignore_single_op_result", action="store_true", default=False, required=False,
        help='<Optional> ignore the single operator detail result, takes effect only when the "-op" exists')
    compare_parser.add_argument(
        '-n', '--topn', dest='topn', type=int, default=ConstManager.DEFAULT_TOP_N,
        help='<Optional> the TopN for the single operator detail result, takes effect '
             'only when the "-op" exists. The value ranges from 1 to 10000. The default value is 20')
    compare_parser.add_argument(
        '-ml', '--max_line', dest='max_line', type=int, default=None,
        help='<Optional> the max line count for the single operator detail result, takes effect '
             'only when the "-op" exists. The default value is 1000000, and it should range '
             'from {} to {}'.format(ConstManager.DETAIL_LINE_COUNT_RANGE_MIN, ConstManager.DETAIL_LINE_COUNT_RANGE_MAX))


def _add_version_argument(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        '-v', '--version', dest='dump_version', choices=[1, 2], type=int, default=2,
        help='<Optional> the version of the dump file, '
             '1 means the protobuf dump file, 2 means the binary dump file, the default value is 2')


def _convert_parser(covert_parser: argparse.ArgumentParser) -> None:
    group = covert_parser.add_mutually_exclusive_group()
    covert_parser.add_argument(
        '-d', '--dump_file', dest='dump_path', default='', type=safe_path_string, required=True,
        help='<Required> the dump file path, supports one file, file list(splits by ",") and directory')
    covert_parser.add_argument('-f', '--format', dest='format', default=None, help='<Optional> the format to transfer')
    covert_parser.add_argument(
        '-s', '--shape', dest='shape', default=None,
        help='<Optional> the shape for format transfer, currently only used for FRACTAL_NZ conversion, '
             'shape format is ([0-9]+,)+[0-9]+, such as 1,3,224,224')
    group.add_argument(
        '-o', '--output_tensor', dest='output', default=None,
        help='<Optional> the index for output, takes effect only when the "-f" exists')
    group.add_argument(
        '-i', '--input_tensor', dest='input', default=None,
        help='<Optional> the index for input, takes effect only when the "-f" exists')
    covert_parser.add_argument(
        '-c', '--custom_script_path', dest='custom_script_path', default=None, type=safe_path_string,
        help='<Optional> the user-defined script path, including format conversion')
    covert_parser.add_argument('-out', '--output', dest='output_path', default='', help='<Optional> the output path')
    _add_version_argument(covert_parser)
    covert_parser.add_argument(
        '-t', '--type', dest='output_file_type', choices=['npy', 'bin', 'msnpy'],
        default='npy',
        help='<Optional> the type of the output file, '
             'npy means the output is saved as numpy format, '
             'bin means the output is saved as binary format, '
             'msnpy means the output is saved as numpy format for MindSpore, '
             'the default value is npy')


def _overflow_parser(overflow_parser: argparse.ArgumentParser) -> None:
    overflow_parser.add_argument(
        '-d', '--dump_path', dest='dump_path', default='', type=safe_path_string, required=True,
        help='<Required> the dump file path')
    overflow_parser.add_argument(
        '-out', '--output', dest='output_path', default='', type=safe_path_string, required=True,
        help='<Optional> the output path')
    overflow_parser.add_argument(
        '-n', '--top_n', dest='top_num', choices=[1, 2, 3, 4, 5], type=int, default=1,
        help='<Optional> the number of overflow ops, first n will be analyzed. the default value is 1')


def _file_compare_parser(file_compare_parser: argparse.ArgumentParser) -> None:
    file_compare_parser.add_argument(
        '-m', '--my_dump_path', dest='my_dump_path', default='', type=safe_path_string,
        help='<Required> my dump path, the data compared with golden data',
        required=True)
    file_compare_parser.add_argument(
        '-g', '--golden_dump_path', dest='golden_dump_path', default='', type=safe_path_string,
        help='<Required> the golden dump path', required=True)
    file_compare_parser.add_argument(
        '-out', '--output', dest='output_path', default='', type=safe_path_string,
        help='<Required> the output path', required=True)


def _check_argument_effect(required_arg: any, options_arg: any, options_arg_str: str, required_arg_str: str) -> None:
    if required_arg is None and options_arg is not None:
        log.print_error_log(
            'The argument %r takes effect only when the "%r" exists.' % (options_arg_str, required_arg_str))
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)


def _check_single_op_argument(args: argparse.Namespace) -> None:
    if args.op_name:
        log.print_error_log('When --mapping or -map exists,the -op parameter is invalid.')
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
    if args.output or args.input:
        log.print_error_log('When --mapping or -map exists,the -i or -o parameter is invalid.')
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
    if not args.fusion_rule_file and not args.quant_fusion_rule_file:
        log.print_error_log('When --mapping or -map exists,there is need to enter the -f or -q parameter.')
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)


def _check_dump_path_exist(dump_path_array: list) -> None:
    for item_path in dump_path_array:
        ret = path_check.check_path_valid(item_path, True)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)


def _check_file_compare_file(args: argparse.Namespace, file_type) -> None:
    for file in [args.my_dump_path, args.golden_dump_path]:
        if not file.endswith(file_type):
            log.print_error_log("[file_compare] The file %r is invalid.Only support %s file." % (file, file_type))
            raise CompareError(CompareError.MSACCUCMP_INVALID_TYPE_ERROR)
        ret = path_check.check_path_valid(file, True, False, path_type=path_check.PathType.File)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)


def _check_file_compare_out(args: argparse.Namespace) -> None:
    ret = path_check.check_output_path_valid(args.output_path, exist=True)
    if ret != CompareError.MSACCUCMP_NONE_ERROR:
        log.print_error_log('[file_compare] The -out parameter: "%r" is invalid!' % args.output_path)
        raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)


def _check_hdf5_file_valid(file_path: str) -> bool:
    """
    Check file is hdf5
    :param file_path: the file path
    :return bool
    """
    return os.path.isfile(os.path.realpath(file_path)) and file_path.endswith(".h5")


def start_compare(args: argparse.Namespace) -> int:
    """
    compare entry.
    """
    if _check_hdf5_file_valid(args.my_dump_path) and _check_hdf5_file_valid(args.golden_dump_path):
        pytorch_compare = PytorchComparison(args)
        pytorch_compare.check_arguments_valid(args)
        check_file_size(args.my_dump_path, ConstManager.ONE_HUNDRED_MB)
        check_file_size(args.golden_dump_path, ConstManager.ONE_HUNDRED_MB)
        ret = check_others_permission(args.my_dump_path)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        ret = check_others_permission(args.golden_dump_path)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        ret = pytorch_compare.compare()
        return ret

    if args.post_process is not None:
        log.print_error_log('param -p only used in pytorch session.')
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)

    if os.path.isfile(os.path.realpath(args.my_dump_path)) and os.path.isfile(os.path.realpath(args.golden_dump_path)):
        ret = check_others_permission(args.my_dump_path)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        ret = check_others_permission(args.golden_dump_path)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        compare = AlgorithmManagerMain(args)
        ret = compare.process()
    elif args.fusion_rule_file != "" and BatchCompare().check_fusion_rule_json_dir(args.fusion_rule_file):
        ret = BatchCompare().compare(args)
    else:
        args = _check_advisor_effect(args)
        compare = VectorComparison(args)
        ret = compare.compare()
    return ret


def _do_cmd() -> int:
    parser = argparse.ArgumentParser(description=MIND_STUDIO_LOGO)
    subparsers = parser.add_subparsers(help='commands')
    compare_parser = subparsers.add_parser(
        'compare', help='Compare network or single op.', description=MIND_STUDIO_LOGO
    )
    covert_parser = subparsers.add_parser(
        'convert', help='Convert my dump data to numpy data or bin data.', description=MIND_STUDIO_LOGO
    )
    overflow_parser = subparsers.add_parser(
        'overflow', help='Analyze the information of the overflow operators.', description=MIND_STUDIO_LOGO
    )
    file_compare_parser = subparsers.add_parser(
        'file_compare', help='Compare two single .npy file.', description=MIND_STUDIO_LOGO
    )

    _compare_parser(compare_parser)
    _convert_parser(covert_parser)
    _overflow_parser(overflow_parser)
    _file_compare_parser(file_compare_parser)

    args = parser.parse_args(sys.argv[1:])
    if len(sys.argv) < 2:
        parser.print_help()
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)

    if sys.argv[1] == 'compare':
        ret = _do_compare(args)
    elif sys.argv[1] == 'convert':
        ret = _do_convert(args)
    elif sys.argv[1] == 'file_compare':
        ret = _do_file_compare(args)
    else:
        ret = _do_overflow(args)

    return ret


def _check_advisor_effect(args):
    if args.advisor and args.range is not None:
        log.print_warn_log('The argument "-advisor" takes no effect when the "-r" or "--range" exists.')
        args.advisor = False
    if args.advisor and args.select is not None:
        log.print_warn_log('The argument "-advisor" takes no effect when the "-s" or "--select" exists.')
        args.advisor = False
    if args.advisor and args.op_name is not None:
        log.print_warn_log('The argument "-advisor" takes no effect when the "-op" exists.')
        args.advisor = False
    if args.advisor:
        log.print_warn_log('The argument "-advisor" will automatically configure "-overflow_detection".')
        args.overflow_detection = True
    return args


def _check_range_effect(args: argparse.Namespace) -> None:
    if not args.fusion_rule_file and args.range is not None:
        log.print_error_log('The argument "-r" or "--range" takes effect only when the "-f" exists.')
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
    if args.op_name and args.range is not None:
        log.print_error_log('The argument "-r" or "--range" exists, the "-op" parameter is invalid.')
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
    if not args.fusion_rule_file and args.select is not None:
        log.print_error_log('The argument "-s" or "--select" takes effect only when the "-f" exists.')
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
    if args.op_name and args.select is not None:
        log.print_error_log('The argument "-s" or "--select" exists, the "-op" parameter is invalid.')
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
    if args.range and args.select is not None:
        log.print_error_log('The argument "-r" and "-s" can not be used at the same time')
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
    if args.max_cmp_size < 0:
        log.print_error_log(
            "Please enter a valid number for max_cmp_size, the max_cmp_size should be"
            " in [0, ∞), now is %s." % args.max_cmp_size)
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)


def _do_compare(args: argparse.Namespace) -> int:
    _check_range_effect(args)
    if args.dump_version == 1:
        log.print_warn_log(
            "The -v argument will be deprecated. when the -v value is 1, it will be processed as 2."
        )
    if args.mapping:
        _check_single_op_argument(args)
        compare = VectorComparison(args)
        ret = compare.compare()
    else:
        _check_argument_effect(args.op_name, args.output, '"-o" or "--output_tensor"', '-op')
        _check_argument_effect(args.op_name, args.input, '"-i" or "--input_tensor"', '-op')
        if '-n' in sys.argv or '--topn' in sys.argv:
            _check_argument_effect(args.op_name, args.topn, '"-n" or "--topn"', '-op')
        if '--ignore_single_op_result' in sys.argv:
            _check_argument_effect(args.op_name, args.ignore_single_op_result, '"--ignore_single_op_result"', '-op')
        if '-ml' in sys.argv or '--max_line' in sys.argv:
            _check_argument_effect(args.op_name, args.max_line, '"-ml" or "--max_line"', '-op')
        dump_path_array = [args.my_dump_path, args.golden_dump_path]
        _check_dump_path_exist(dump_path_array)
        ret = start_compare(args)

    return ret


def _do_convert(args: argparse.Namespace) -> int:
    _check_argument_effect(args.format, args.output, '"-o" or "--output_tensor"', '-f')
    _check_argument_effect(args.format, args.input, '"-i" or "--input_tensor"', '-f')
    _check_argument_effect(args.format, args.shape, '"-s" or "--shape"', '-f')
    _check_argument_effect(args.format, args.custom_script_path, '"-c" or "--custom_script_path"', '-f')
    if args.dump_version == 1:
        log.print_warn_log(
            "The -v argument will be deprecated. when the -v value is 1, it will be processed as 2."
        )
    abs_dump_path = os.path.abspath(args.dump_path)
    if os.path.isdir(abs_dump_path) and not os.listdir(abs_dump_path):
        log.print_error_log('The dump path is empty.')
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
    if args.format:
        if os.path.isdir(abs_dump_path):
            log.print_warn_log(
                'The dump path is a directory. If the -o, -i and -s arguments exist, these arguments will be ignored.'
            )
        conversion = FormatConversionMain(args)
        ret = conversion.convert_format()
    else:
        ret = DumpDataParser(args).parse_dump_data()

    return ret


def _do_file_compare(args: argparse.Namespace) -> int:
    _check_file_compare_file(args, ConstManager.NPY_SUFFIX)
    _check_file_compare_out(args)
    args.custom_script_path = ""
    args.algorithm = ConstManager.FILE_CMP_SUPPORTED_ALGORITHM
    args.algorithm_options = ""
    compare = AlgorithmManagerMain(args)
    ret = compare.process(save_result=True)

    return ret


def _do_overflow(args: argparse.Namespace) -> int:
    overflow_analyse = OverflowAnalyse(args)
    ret = overflow_analyse.check_argument(args)
    if ret == CompareError.MSACCUCMP_NONE_ERROR:
        return overflow_analyse.analyse()
    return ret


def _root_privilege_warning():
    if os.getuid() == 0:
        log.print_warn_log(
            "msaccucmp is being run as root. "
            "To avoid security risks, it is recommended to switch to a regular user to run it."
        )


def main() -> None:
    """
    parse argument and run command
    :return:
    """
    start = time.time()
    with file_utils.UmaskWrapper():
        try:
            _root_privilege_warning()
            ret = _do_cmd()
        except CompareError as err:
            ret = err.code
        except Exception as base_err:
            log.print_error_log(f'Basic error running {sys.argv[0]}: {base_err}')
            sys.exit(1)
    end = time.time()
    log.print_info_log('The command was completed and took %d seconds.' % (end - start))
    sys.exit(ret)


if __name__ == '__main__':
    main()
